# encoding = utf-8

import os
import sys
import requests
import json
import urllib3
from urllib.parse import urlparse, urlencode, urlunparse, parse_qs
import time
import hashlib
from splunklib.modularinput.event import Event, ET
from splunklib.modularinput.event_writer import EventWriter
from trackme_libs import trackme_reqinfo, trackme_idx_for_tenant, trackme_state_event
import splunklib.client as client

import smtplib
import ssl
import random
import string
from email.message import EmailMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import pygal
from pygal.style import DarkStyle
from io import BytesIO
import base64

# Disable urllib3 warnings for SSL
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Set splunkhome
splunkhome = os.environ["SPLUNK_HOME"]

# Append libs
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

from trackme_libs import run_splunk_search


def generate_message_id():
    """
    Function to generate a message ID

    Args:
        None

    Returns:
        A message ID
    """

    # Create a random message ID with your domain
    random_hash = hashlib.sha256(str(time.time()).encode()).hexdigest()[:16]
    return random_hash


def generate_drilldown_link(
    parsed_url, drilldown_root_uri, tenant_id, object_category, object_value
):
    """
    Function to generate the drilldown link for a given object

    Args:
        drilldown_root_uri: The root URI to build the drilldown link
        object: The object to build the drilldown link for

    Returns:
        The drilldown link
    """

    # Extract the server name (hostname) and server port
    server_protocol = parsed_url.scheme
    server_name = parsed_url.hostname
    server_port = parsed_url.port

    # extract the component suffix as splk-<component>
    component_suffix = object_category.split("-")[-1]

    # Base URL without query parameters
    base_url = f"{server_protocol}://{server_name}"
    if not (server_protocol == "https" and server_port == 443) and not (
        server_protocol == "http" and server_port == 80
    ):
        base_url += f":{server_port}"

    # Path and query parameters
    path = "/app/trackme/trackMe_home"
    query_params = {
        "tenant_id": tenant_id,
        "component": component_suffix,
        "object": object_value,
    }

    # Construct the full URL
    if drilldown_root_uri:
        # Parse the drilldown_root_uri
        parsed_uri = urlparse(drilldown_root_uri)
        base_url = f"{parsed_uri.scheme}://{parsed_uri.netloc}"

        # Preserve the path from the drilldown_root_uri
        root_path = parsed_uri.path.rstrip("/")  # Strip trailing slash for consistency
        full_path = f"{root_path}{path}"  # Combine root path with the specific app path

        # Adjust query parameters if needed
        _, adjusted_query_params = adjust_query_params(
            drilldown_root_uri, query_params.copy()
        )
        drilldown_link = f"{base_url}{full_path}?{urlencode(adjusted_query_params)}"
    else:
        base_url = f"{server_protocol}://{server_name}"
        if not (server_protocol == "https" and server_port == 443) and not (
            server_protocol == "http" and server_port == 80
        ):
            base_url += f":{server_port}"
        full_path = path  # Default to the base path if no drilldown_root_uri provided
        drilldown_link = f"{base_url}{full_path}?{urlencode(query_params)}"

    return drilldown_link


def adjust_query_params(url, query_params):
    """
    Function to adjust the query parameters for the drilldown link

    Args:
        url: The URL to adjust the query parameters for
        query_params: The query parameters to adjust

    Returns:
        The adjusted query parameters
    """

    parsed_url = urlparse(url)
    query_dict = parse_qs(parsed_url.query)
    if "tenant_id" in query_dict:
        query_params.pop("tenant_id", None)
    separator = "&" if parsed_url.query else "?"
    return separator, query_params


def normalize_message(msg):
    if isinstance(msg, list):
        return " ".join(str(x).strip() for x in msg)
    elif isinstance(msg, str):
        return msg.strip()
    return str(msg)


def send_trackme_email(
    helper,
    config,
    message_id,
    to_address,
    subject,
    core_message,
    in_reply_to=None,
    reference_chain=None,
    format="html",
    tenant_id=None,
    alias=None,
    object=None,
    object_state=None,
    object_category=None,
    object_id=None,
    anomaly_reason=None,
    status_message=None,
    message_source=None,
    message_source_id=None,
    transition_message=None,
    environment_name=None,
    drilldown_link=None,
    incident_id=None,
    detection_time=None,
    chart_svg_base64_list=None,
):
    """
    Function to send an email

    Args:
        config: The email account settings
        message_id: The message ID
        to_address: The address to send the email to
        subject: The subject of the email
        body: The body of the email

    Notes:
        - For Splunk Cloud vetting purposes, email security is mandatory except for the localhost SMTP server

    """

    # check target, if not localhost for the local MTA, then security is mandatory, raise an error if not configured
    email_security = config.get("email_security", None)
    if not email_security:
        raise RuntimeError("Email security is mandatory for external email delivery")
    elif email_security not in ["ssl", "tls"]:
        raise RuntimeError(
            "Invalid email security configuration, Email security must be ssl or tls"
        )

    # check charts
    chart_svg_base64_list = chart_svg_base64_list or []

    # check messages
    normalized_status = normalize_message(status_message)
    normalized_core = normalize_message(core_message)

    # Decide if we should include status_message separately
    include_status_message = normalized_status and normalized_status != normalized_core

    # Construct the message
    msg = MIMEMultipart("alternative")
    msg["Subject"] = subject
    msg["From"] = config["sender_email"]
    msg["To"] = to_address

    # Generate domain for Message-ID
    domain = (
        config["sender_email"].split("@")[-1]
        if "@" in config["sender_email"]
        else "trackme"
    )

    msg["Message-ID"] = f"<{message_id}@{domain}>"

    if in_reply_to:
        msg["In-Reply-To"] = f"<{in_reply_to}@{domain}>"

    if reference_chain:
        msg["References"] = " ".join(f"<{mid}@{domain}>" for mid in reference_chain)

    # Prepare footer
    footer = config.get(
        "email_footer",
        "This is an automated email, please do not reply directly to this email.",
    )

    # Construct plain text version
    plain_body = f"""Hello,

The TrackMe entity:

- Incident ID: {incident_id}
- Detection Time: {detection_time}
- Environment: {environment_name}
- Tenant: {tenant_id}
- Alias: {alias}
- Object: {object}
- Category: {object_category}
- Object ID: {object_id}
- Drilldown Link: {drilldown_link}
- Object State: {object_state}
- Anomaly Reason: {anomaly_reason}
- Status Message: {status_message}

{transition_message}

{core_message}

Regards,
TrackMe Alerting System

{footer}
"""
    part1 = MIMEText(plain_body, "plain")

    # Construct HTML version if needed
    if format == "html":
        helper.log_info(f"Sending HTML email to {to_address}")

        icon = (
            "🚨"
            if object_state == "red"
            else (
                "🟠"
                if object_state == "orange"
                else "🔵" if object_state == "blue" else "✅"
            )
        )
        state_color = (
            "#d9534f"
            if object_state == "red"
            else (
                "#f0ad4e"
                if object_state == "orange"
                else "#007bff" if object_state == "blue" else "#5cb85c"
            )
        )

        html_body = f"""
        <html>
        <body style="font-family: Arial, sans-serif; font-size: 14px; color: #333;">
            <div style="border: 1px solid #ccc; padding: 16px; border-radius: 6px; background-color: #fefefe;">
            <h2 style="color: {state_color};">{icon} TrackMe Notification</h2>
            <ul style="list-style-type: none; padding-left: 0;">
                <li><strong>Environment:</strong> {environment_name}</li>
                <li><strong>Detection Time:</strong> {detection_time}</li>
                <li><strong>Tenant:</strong> {tenant_id}</li>
                <li><strong>Alias:</strong> {alias}</li>
                <li><strong>Object:</strong> {object}</li>
                <li><a href="{drilldown_link}">Drilldown Link</a></li>
                <br />
                <li><strong>Status:</strong> <span style="color: {state_color}; font-weight: bold;">{object_state}</span></li>
                <br />
                <li><strong>Incident ID:</strong> {incident_id}</li>
            </ul>
            <hr style="margin: 20px 0;">
            <div style="background-color: #2b2b2b; color: #f1f1f1; padding: 12px; border-radius: 6px; font-family: monospace;">
                <p style="margin: 0 0 10px 0;"><strong>📋 Detailed Information:</strong></p>
                <p style="margin-bottom: 10px;">{transition_message}</p>
                <ul style="list-style-type: disc; padding-left: 20px;">
                <li><strong>Alias:</strong> {alias}</li>
                <li><strong>Object:</strong> {object}</li>
                <li><strong>Object ID:</strong> {object_id}</li>
                <li><strong>Category:</strong> {object_category}</li>
                <li><strong>Anomaly Reason:</strong> {anomaly_reason}</li>
                {f'<li><strong>Status Message:</strong> {status_message}</li>' if include_status_message else ''}
                <li><strong>Message:</strong> {core_message}</li>
                <li><strong>Message source:</strong> {message_source}</li>
                <li><strong>Message source ID:</strong> {message_source_id}</li>
                </ul>
            </div>
        """

        # Inject charts if present
        if chart_svg_base64_list:
            html_body += """
            <hr style="margin: 30px 0;">
            <h3 style="margin-bottom: 10px;">📈 Performance Charts (24h)</h3>
            """

            # Group charts 3 per row
            for i in range(0, len(chart_svg_base64_list), 3):
                charts_row = chart_svg_base64_list[i : i + 3]
                html_body += (
                    '<div style="display: flex; gap: 10px; margin-bottom: 10px;">'
                )

                # Add real charts
                for svg_b64 in charts_row:
                    html_body += f"""
                    <div style="flex: 1; min-width: 250px;">
                        <img src="data:image/svg+xml;base64,{svg_b64}" alt="Chart" style="max-width: 100%; height: auto; border: 1px solid #ccc; border-radius: 8px;" />
                    </div>
                    """

                # Add empty placeholders to make up to 3
                missing = 3 - len(charts_row)
                for _ in range(missing):
                    html_body += """
                    <div style="flex: 1; min-width: 250px;"></div>
                    """

                html_body += "</div>"

        # Close the HTML
        html_body += f"""
            <p style="margin-top: 20px; font-size: 12px; color: #888;">{footer}</p>
            </div>
        </body>
        </html>
        """

        part2 = MIMEText(html_body, "html")
        msg.attach(part1)
        msg.attach(part2)
    else:
        msg.attach(part1)

    # Extract server & port
    server_parts = config["email_server"].split(":")
    smtp_server = server_parts[0]
    smtp_port = int(server_parts[1])
    context = ssl.create_default_context()

    try:
        if config["email_security"] == "ssl":
            with smtplib.SMTP_SSL(smtp_server, smtp_port, context=context) as server:
                if config.get("email_username") and config.get("email_password"):
                    server.login(config["email_username"], config["email_password"])
                server.send_message(msg)
        else:
            with smtplib.SMTP(smtp_server, smtp_port) as server:
                if config["email_security"] == "tls":
                    server.starttls(context=context)
                if config.get("email_username") and config.get("email_password"):
                    server.login(config["email_username"], config["email_password"])
                server.send_message(msg)

        return message_id

    except Exception as e:
        raise RuntimeError(f"Failed to send email: {e}")


def ingest_stateful_alert_event(
    helper, session_key, server_uri, tenant_id, index, record
):
    """
    Function to ingest a stateful alert event into the TrackMe summary index

    Args:
        session_key: The session key
        server_uri: The server URI
        tenant_id: The tenant ID
        object_category: The object category
        object_value: The object value
    """

    # generate event
    try:
        trackme_state_event(
            session_key=session_key,
            splunkd_uri=server_uri,
            tenant_id=tenant_id,
            index=index,
            sourcetype="trackme:stateful_alerts",
            source=f"trackme_stateful_alerts_{tenant_id}",
            record=record,
        )
        helper.log_info(
            f'tenant_id="{tenant_id}", object_category="{record.get("object_category")}", object="{record.get("object")}", Stateful alert event created successfully, response="{json.dumps(record, indent=2)}"'
        )
        return True

    except Exception as e:
        helper.log_error(
            f'tenant_id="{tenant_id}", object_category="{record.get("object_category")}", object="{record.get("object")}", Stateful alert event creation failure, response="{json.dumps(record, indent=2)}", exception="{str(e)}"'
        )
        raise RuntimeError(f"Failed to ingest stateful alert event: {e}")


def get_keyid_from_main_kvstore(helper, collection, object):
    """
    Function to get the keyid value from the main KVstore collection and for a given object

    Args:
        collection: The collection to query
        object: The object to query

    Returns:
        The keyid value
    """

    # Define the KV query
    query_string = {
        "object": object,
    }

    # Get the current record
    try:
        kvrecord = collection.data.query(query=json.dumps(query_string))[0]
        key = kvrecord.get("_key")

    except Exception as e:
        key = None

    if kvrecord:
        return kvrecord.get("_key")
    else:
        return None


def get_object_state_from_main_kvstore(helper, collection, object_id):
    """
    Function to get the object_state value from the main KVstore collection and for a given object_id

    Args:
        collection: The collection to query
        object_id: The object ID to query

    Returns:
        The object_state value, anomaly_reason and status_message
    """

    # Define the KV query
    query_string = {
        "_key": object_id,
    }

    # Get the current record
    try:
        kvrecord = collection.data.query(query=json.dumps(query_string))[0]
        helper.log_debug(f"main collection kvrecord: {json.dumps(kvrecord, indent=2)}")
        key = kvrecord.get("_key")

    except Exception as e:
        key = None

    if kvrecord:

        object_state = kvrecord.get("object_state")
        anomaly_reason = kvrecord.get("anomaly_reason")
        status_message = kvrecord.get("status_message")

        return object_state, anomaly_reason, status_message
    else:
        return "green", None, None


def get_stateful_record_for_message_id(collection, message_id):
    """
    Function to get the stateful record from the KVstore for a given message_id

    Args:
        collection: The collection to query
        message_id: The message ID to query

    Returns:
        The stateful record
    """

    # Define the KV query
    query_string = {
        "message_id": message_id,
    }

    # Get the current record
    try:
        kvrecord = collection.data.query(query=json.dumps(query_string))[0]
        key = kvrecord.get("_key")

    except Exception as e:
        key = None

    return kvrecord


def get_stateful_records_for_object_id(helper, collection, object_id):
    """
    Function to get all stateful records for a given object_id

    Args:
        collection: The collection to query
        object_id: The object ID to query

    Returns:
        The last stateful record for the given object_id

    """

    # Define the KV query
    query_string = {
        "object_id": object_id,
    }

    # Get all records
    try:
        kvrecords = collection.data.query(query=json.dumps(query_string))
        helper.log_debug(f"kvrecords: {json.dumps(kvrecords, indent=2)}")
    except Exception as e:
        raise RuntimeError(f"Failed to get stateful records: {e}")

    if kvrecords:
        # select the newer record based on mtime (epochtime)
        kvrecords.sort(key=lambda x: x.get("mtime"), reverse=True)
        last_kvrecord = kvrecords[0]
        helper.log_debug(f"last_kvrecord: {json.dumps(last_kvrecord, indent=2)}")

        # get the alert_status from the last_kvrecord, if it closed, return None as this must be considered as a new thread
        alert_status = last_kvrecord.get("alert_status", "opened")
        if alert_status == "closed":
            return None
        else:
            return last_kvrecord

    else:
        return None


def insert_stateful_record(collection, record):
    """
    Function to insert a new stateful record into the KVstore

    Args:
        collection: The collection to insert the record into
        record: The record to insert

    Returns:
        The inserted record

    """

    try:
        collection.data.insert(json.dumps(record))
    except Exception as e:
        raise RuntimeError(f"Failed to insert stateful record: {e}")


def update_stateful_record(collection, record):
    """
    Function to update an existing stateful record in the KVstore

    Args:
        collection: The collection to update the record in
        record: The record to update

    Returns:
        The updated record

    """

    try:
        collection.data.update(record.get("_key"), json.dumps(record))
    except Exception as e:
        raise RuntimeError(f"Failed to update stateful record: {e}")


def flx_get_metrics_catalog_for_object_id(helper, service, tenant_id, object_id):
    """
    Function to get the metrics catalog for a given object_id

    Args:
        helper: The helper object
        service: The service object
        tenant_id: The tenant ID
        object_id: The object ID

    Returns:
        The metrics catalog (list)

    """

    # Dynamically discover additional flx metrics
    metrics_catalog = []
    try:
        mcatalog_query = f"""| mcatalog values(metric_name) as metrics where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" metric_name="trackme.splk.flx.*" metric_name!="trackme.splk.flx.status"
        | mvexpand metrics | rex field=metrics "trackme\.splk\.flx\.(?<metrics>.*)" | stats values(metrics) as metrics """
        search_kwargs = {
            "output_mode": "json",
            "count": 0,
            "earliest_time": "-24h@h",
            "latest_time": "now",
        }
        reader = run_splunk_search(service, mcatalog_query, search_kwargs, 24, 5)
        for item in reader:
            if isinstance(item, dict):
                metrics_result = item.get("metrics", [])
                break  # only one result is expected

        # ensure metrics_result is a list
        if metrics_result:
            if not isinstance(metrics_result, list):
                metrics_result = [metrics_result]

        # return the metrics_result
        return metrics_result

    except Exception as e:
        helper.log_error(f"Error retrieving flx metrics for object_id={object_id}: {e}")
        return None


def wlk_get_metrics_catalog_for_object_id(helper, service, tenant_id, object_id):
    """
    Function to get the metrics catalog for a given object_id

    Args:
        helper: The helper object
        service: The service object
        tenant_id: The tenant ID
        object_id: The object ID

    Returns:
        The metrics catalog (list)

    """

    # Dynamically discover additional flx metrics
    metrics_catalog = []
    try:
        mcatalog_query = f"""| mcatalog values(metric_name) as metrics where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" metric_name="trackme.splk.wlk.*" 
        | mvexpand metrics | rex field=metrics "trackme\.splk\.wlk\.(?<metrics>.*)" | stats values(metrics) as metrics """
        search_kwargs = {
            "output_mode": "json",
            "count": 0,
            "earliest_time": "-24h@h",
            "latest_time": "now",
        }
        reader = run_splunk_search(service, mcatalog_query, search_kwargs, 24, 5)
        for item in reader:
            if isinstance(item, dict):
                metrics_result = item.get("metrics", [])
                break  # only one result is expected

        # ensure metrics_result is a list
        if metrics_result:
            if not isinstance(metrics_result, list):
                metrics_result = [metrics_result]

        # return the metrics_result
        return metrics_result

    except Exception as e:
        helper.log_error(f"Error retrieving wlk metrics for object_id={object_id}: {e}")
        return None


def get_mlmodels_in_anomaly_from_kvstore(
    helper, service, tenant_id, component, object, object_id
):
    """
    Function to get the mlmodels_in_anomaly from the KVstore for a given object_id

    Args:
        helper: The helper object
        service: The service object
        tenant_id: The tenant ID
        component: The component
        object: The object
        object_id: The object ID

    Returns:
        The mlmodels_in_anomaly

    """

    # Get the current record
    try:

        # connect to the main tenant KVstore collection
        collection_name = (
            f"kv_trackme_{component}_outliers_entity_data_tenant_{tenant_id}"
        )
        collection = service.kvstore[collection_name]

        # Define the KV query
        query_string = {
            "_key": object_id,
        }

        kvrecord = collection.data.query(query=json.dumps(query_string))[0]
        helper.log_debug(f"main collection kvrecord: {json.dumps(kvrecord, indent=2)}")
        key = kvrecord.get("_key")
        models_in_anomaly = kvrecord.get("models_in_anomaly", [])

        # if we have a result, ensure this is a list and convert otherwise
        if models_in_anomaly:
            if not isinstance(models_in_anomaly, list):
                models_in_anomaly = [models_in_anomaly]

        return models_in_anomaly

    except Exception as e:
        return None


def generate_chart_base64(
    title, x_labels, series_dict, chart_type="line", theme="dark"
):
    """
    Function to generate a chart base64

    Args:
        title: The title of the chart
        x_labels: The x labels of the chart
        series_dict: The series dictionary of the chart
        chart_type: The type of the chart

    Returns:
        The chart base64
    """

    if theme == "light":
        custom_style = pygal.style.LightStyle
    else:  # fallback to dark
        custom_style = pygal.style.DarkStyle(
            background="transparent",
            plot_background="#1e1e1e",
            foreground="#ffffff",
            foreground_strong="#ffffff",
            foreground_subtle="#aaaaaa",
            label_font_size=10,
            major_label_font_size=10,
            legend_font_size=12,
            title_font_size=14,
        )

    # Detect if this chart should be scaled from 0 to 100 (for percentage metrics)
    percentage_scale = any(
        "pct" in k.lower() or "percent" in k.lower() for k in series_dict
    )

    # Define common chart kwargs
    chart_kwargs = {
        "height": 250,
        "width": 500,
        "show_legend": True,
        "explicit_size": True,
        "x_labels": [],
        "show_x_labels": False,
        "show_minor_x_labels": False,
        "legend_at_bottom": True,
        "style": custom_style,
        "show_dots": False,
    }

    if percentage_scale:
        chart_kwargs["range"] = (0, 100)

    # Choose chart type
    if chart_type == "bar":
        chart = pygal.Bar(**chart_kwargs)
    else:  # fallback to line
        chart = pygal.Line(**chart_kwargs)

    chart.title = title
    for label, values in series_dict.items():
        chart.add(label, values)
    svg_data = chart.render()
    return base64.b64encode(svg_data).decode("utf-8")


def build_latency_chart(search_results, theme_charts=None):
    """
    Function to build a latency chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The latency chart
    """

    try:
        timestamps, avg, latest, perc95 = [], [], [], []
        for row in search_results:
            timestamps.append(row.get("_time"))
            avg.append(float(row.get("avg_latency_5m", 0)))
            latest.append(float(row.get("latest_latency_5m", 0)))
            perc95.append(float(row.get("perc95_latency_5m", 0)))
        return generate_chart_base64(
            "Latency Metrics - Last 24h",
            timestamps,
            {
                "Avg Latency (5m)": avg,
                "Latest Latency (5m)": latest,
                "95th Perc Latency (5m)": perc95,
            },
            theme=theme_charts,
        )
    except Exception:
        return None


def build_delay_chart(search_results, theme_charts=None):
    """
    Function to build a delay chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The delay chart
    """

    try:
        timestamps, delays = [], []
        for row in search_results:
            timestamps.append(row.get("_time"))
            delays.append(float(row.get("lag_event_sec", 0)))
        return generate_chart_base64(
            "Delay Metrics - Last 24h",
            timestamps,
            {
                "Event Lag (sec)": delays,
            },
            theme=theme_charts,
        )
    except Exception:
        return None


def build_volume_chart(search_results, theme_charts=None):
    """
    Function to build a volume chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The volume chart
    """

    try:
        timestamps, avg, latest, perc95 = [], [], [], []
        for row in search_results:
            timestamps.append(row.get("_time"))
            avg.append(float(row.get("avg_eventcount_5m", 0)))
            latest.append(float(row.get("latest_eventcount_5m", 0)))
            perc95.append(float(row.get("perc95_eventcount_5m", 0)))
        return generate_chart_base64(
            "Volume Metrics - Last 24h",
            timestamps,
            {
                "Avg Events (5m)": avg,
                "Latest Events (5m)": latest,
                "95th Perc Events (5m)": perc95,
            },
            theme=theme_charts,
        )
    except Exception:
        return None


def build_hosts_dcount_chart(search_results, theme_charts=None):
    """
    Function to build a hosts dcount chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The hosts dcount chart
    """

    try:
        timestamps = []
        global_dcount_host = []
        avg_dcount_host_5m = []
        latest_dcount_host_5m = []
        perc95_dcount_host_5m = []

        for row in search_results:
            timestamps.append(row.get("_time"))
            global_dcount_host.append(float(row.get("global_dcount_host", 0)))
            avg_dcount_host_5m.append(float(row.get("avg_dcount_host_5m", 0)))
            latest_dcount_host_5m.append(float(row.get("latest_dcount_host_5m", 0)))
            perc95_dcount_host_5m.append(float(row.get("perc95_dcount_host_5m", 0)))

        return generate_chart_base64(
            "Hosts DCount Metrics - Last 24h",
            timestamps,
            {
                "Global DCount (Host)": global_dcount_host,
                "Avg DCount (5m)": avg_dcount_host_5m,
                "Latest DCount (5m)": latest_dcount_host_5m,
                "95th Perc DCount (5m)": perc95_dcount_host_5m,
            },
            theme=theme_charts,
        )
    except Exception as e:
        return None


def build_flx_status_chart(search_results, theme_charts=None):
    """
    Function to build a flx status chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The flx status chart
    """

    try:
        timestamps = []
        status = []

        for row in search_results:
            timestamps.append(row.get("_time"))
            status.append(float(row.get("status", 0)))

        return generate_chart_base64(
            "FLX Status - Last 24h",
            timestamps,
            {
                "Status": status,
            },
            theme=theme_charts,
        )
    except Exception as e:
        return None


def build_ml_outliers_chart(search_results, theme_charts=None):
    """
    Function to build a ml outliers chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The ml outliers chart
    """

    try:
        timestamps = []
        kpi_metric_value = []
        LowerBound = []
        UpperBound = []
        for row in search_results:
            timestamps.append(row.get("_time"))
            kpi_metric_value.append(float(row.get("kpi_metric_value", 0)))
            LowerBound.append(float(row.get("LowerBound", 0)))
            UpperBound.append(float(row.get("UpperBound", 0)))

        return generate_chart_base64(
            "Machine Learning Outliers - Last 24h",
            timestamps,
            {
                "KPI Metric Value": kpi_metric_value,
                "Lower Bound": LowerBound,
                "Upper Bound": UpperBound,
            },
            theme=theme_charts,
        )
    except Exception as e:
        return None


def build_data_sampling_anomaly_chart(search_results, theme_charts=None):
    """
    Function to build a data sampling anomaly chart

    Args:
        search_results: The search results
        theme_charts: The email charts theme

    Returns:
        The data sampling anomaly chart
    """

    try:
        if not search_results:
            return None

        timestamps = []
        series = {}

        for row in search_results:
            ts = row.get("_time")
            if ts not in timestamps:
                timestamps.append(ts)

            for k, v in row.items():
                if k in ("_time", "_span"):
                    continue
                if k not in series:
                    series[k] = []
                series[k].append(float(v))

        return generate_chart_base64(
            "Model Sampling Match % - Last 24h",
            timestamps,
            series,
            chart_type="bar",
            theme=theme_charts,
        )

    except Exception as e:
        return None


def build_requested_charts(
    helper, chart_builders, chart_types, chart_results_map, theme_charts=None
):
    """
    Function to build requested charts

    Args:
        helper: The helper object
        chart_builders: The chart builders
        chart_types: The chart types
        chart_results_map: The chart results map
        theme_charts: The email charts theme

    Returns:
        The list of SVGs for the requested charts

    """

    svg_charts = []
    for chart_type, model_id in chart_types:

        if chart_type in ("flx_metric_group", "wlk_metric_group"):
            for metric_name in model_id:  # model_id is actually the list of metrics
                metric_data = []
                for row in chart_results_map.get(chart_type, []):
                    if metric_name in row:
                        metric_data.append(
                            {"_time": row.get("_time"), metric_name: row[metric_name]}
                        )
                builder = chart_builders.get(
                    "flx_metric" if "flx" in chart_type else "wlk_metric"
                )
                svg = builder(helper, metric_data, metric_name)
                if svg:
                    svg_charts.append(svg)

        else:
            builder = chart_builders.get(chart_type)
            if not builder:
                helper.log_warn(f"No builder for chart type: {chart_type}")
                continue
            search_key = f"{chart_type}:{model_id}" if model_id else chart_type
            svg = builder(
                chart_results_map.get(search_key, []),
                theme_charts=theme_charts,
            )

        if svg:
            svg_charts.append(svg)

    return svg_charts


def build_flx_generic_metric_chart(
    helper, search_results, metric_name="", theme_charts=None
):
    """
    Function to build a flx generic metric chart

    Args:
        helper: The helper object
        search_results: The search results
        metric_name: The metric name
        theme_charts: The email charts theme

    Returns:
        The flx generic metric chart
    """

    try:
        if not search_results:
            return None

        if search_results:
            first_row_keys = list(search_results[0].keys())
            helper.log_info(
                f"[FLX Chart] First row keys for metric '{metric_name}': {first_row_keys}"
            )
        else:
            helper.log_info(f"[FLX Chart] No results for metric '{metric_name}'")

        timestamps = []
        series = {}
        for row in search_results:
            ts = row.get("_time")
            if ts not in timestamps:
                timestamps.append(ts)
            for k, v in row.items():
                if k in ("_time", "_span"):
                    continue
                if k not in series:
                    series[k] = []
                series[k].append(float(v))
        title = f"FLX Metric: {metric_name} - Last 24h"
        return generate_chart_base64(title, timestamps, series, theme=theme_charts)
    except Exception as e:
        return None


def build_wlk_generic_metric_chart(
    helper, search_results, metric_name="", theme_charts=None
):
    """
    Function to build a wlk generic metric chart

    Args:
        helper: The helper object
        search_results: The search results
        metric_name: The metric name
        theme_charts: The email charts theme

    Returns:
        The wlk generic metric chart
    """

    try:
        if not search_results:
            return None

        if search_results:
            first_row_keys = list(search_results[0].keys())
            helper.log_info(
                f"[WLK Chart] First row keys for metric '{metric_name}': {first_row_keys}"
            )
        else:
            helper.log_info(f"[WLK Chart] No results for metric '{metric_name}'")

        timestamps = []
        series = {}
        for row in search_results:
            ts = row.get("_time")
            if ts not in timestamps:
                timestamps.append(ts)
            for k, v in row.items():
                if k in ("_time", "_span"):
                    continue
                if k not in series:
                    series[k] = []
                series[k].append(float(v))
        title = f"WLK Metric: {metric_name} - Last 24h"
        return generate_chart_base64(title, timestamps, series, theme=theme_charts)
    except Exception as e:
        return None


def get_chart_search(
    chart_type=None,
    tenant_id=None,
    object_category=None,
    object=None,
    keyid=None,
    model_id=None,
    metric_list=None,
):
    """
    Function to get the chart search

    Args:
        chart_type: The chart type
        tenant_id: The tenant ID
        object_category: The object category
        object: The object
        keyid: The keyid
        model_id: The model ID
        metric_list: The metric list

    Returns:
        The chart search
    """

    # get component from object_category (splk-<component>)
    component = object_category.split("-")[1]

    if chart_type == "latency":
        return f"""| mstats avg(trackme.splk.feeds.avg_latency_5m) as avg_latency_5m, avg(trackme.splk.feeds.latest_latency_5m) as latest_latency_5m, avg(trackme.splk.feeds.perc95_latency_5m) as perc95_latency_5m where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="{object_category}" object="{object}" by object span=5m
        | timechart avg(avg_latency_5m) as avg_latency_5m, avg(latest_latency_5m) as latest_latency_5m, avg(perc95_latency_5m) as perc95_latency_5m"""

    elif chart_type == "delay":
        return f"""| mstats avg(trackme.splk.feeds.lag_event_sec) as lag_event_sec where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="{object_category}" object="{object}" by object span=5m
        | timechart avg(lag_event_sec) as lag_event_sec"""

    elif chart_type == "volume":
        return f"""| mstats avg(trackme.splk.feeds.avg_eventcount_5m) as avg_eventcount_5m, avg(trackme.splk.feeds.latest_eventcount_5m) as latest_eventcount_5m, avg(trackme.splk.feeds.perc95_eventcount_5m) as perc95_eventcount_5m where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="{object_category}" object="{object}" by object span=5m
        | timechart avg(avg_eventcount_5m) as avg_eventcount_5m, avg(latest_eventcount_5m) as latest_eventcount_5m, avg(perc95_eventcount_5m) as perc95_eventcount_5m"""

    elif chart_type == "hosts_dcount":
        return f"""| mstats avg(trackme.splk.feeds.global_dcount_host) as global_dcount_host, avg(trackme.splk.feeds.avg_dcount_host_5m) as avg_dcount_host_5m, avg(trackme.splk.feeds.latest_dcount_host_5m) as latest_dcount_host_5m, avg(trackme.splk.feeds.perc95_dcount_host_5m) as perc95_dcount_host_5m where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="{object_category}" object="{object}" by object span=5m
        | timechart span=5m avg(global_dcount_host) as global_dcount_host, avg(avg_dcount_host_5m) as avg_dcount_host_5m, avg(latest_dcount_host_5m) as latest_dcount_host_5m, avg(perc95_dcount_host_5m) as perc95_dcount_host_5m"""

    elif chart_type == "data_sampling_anomaly":
        return f"""| mstats latest(trackme.splk_dsm.sampling.model_pct_match) as model_pct_match where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="{object_category}" object="{object}" by model_name span=5m
        | timechart avg(model_pct_match) as model_pct_match by model_name"""

    elif chart_type == "flx_status":
        return f"""| mstats min(trackme.splk.flx.status) as status where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_id="{keyid}" span=5m
        | timechart min(status) as status"""

    elif chart_type == "ml_outliers":
        return f"""| trackmesplkoutliersrender tenant_id="{tenant_id}" component="{component}" object="{object}" model_id="{model_id}"
        | table _time, kpi_metric_value, LowerBound, UpperBound"""

    elif chart_type == "flx_metric_group" and metric_list:
        mstats_parts = [
            f"max(trackme.splk.flx.{metric}) as {metric}" for metric in metric_list
        ]
        mstats_fields = ", ".join(mstats_parts)
        return f"""| mstats {mstats_fields} where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_id="{keyid}" span=5m
        | timechart span=5m {', '.join([f'max({m}) as {m}' for m in metric_list])}"""

    elif chart_type == "wlk_metric_group" and metric_list:
        mstats_parts = [
            f"max(trackme.splk.wlk.{metric}) as {metric}" for metric in metric_list
        ]
        mstats_fields = ", ".join(mstats_parts)
        return f"""| mstats {mstats_fields} where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_id="{keyid}" span=5m
        | timechart span=5m {', '.join([f'max({m}) as {m}' for m in metric_list])}"""

    return None


def process_event(helper, *args, **kwargs):
    """
    Function to process the event

    Args:
        helper: The helper object

    Returns:
        The return code
    """

    helper.log_info("Alert action trackme_stateful_alert started.")

    # Retrieve the session_key
    helper.log_debug("Get session_key.")
    session_key = helper.session_key

    # Retrieve the server_uri
    helper.log_debug("Get server_uri.")
    server_uri = helper.settings["server_uri"]

    # results_link
    helper.log_debug("Get results_link.")
    results_link = helper.settings["results_link"]

    # from results_link, extract server_name and server_port
    parsed_url = urlparse(results_link)

    # Get request info and set logging level
    reqinfo = trackme_reqinfo(session_key, server_uri)

    # Get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=session_key,
        timeout=300,
    )

    # Build header and target
    header = {
        "Authorization": f"Splunk {session_key}",
        "Content-Type": "application/json",
    }

    # get the delivery_target
    delivery_target = helper.get_param("delivery_target")

    # email account
    email_account = helper.get_param("email_account")

    # Option to include chart in emails (0 or 1)
    generate_charts = int(helper.get_param("generate_charts") or 0)

    # email charts theme
    theme_charts = helper.get_param("theme_charts")

    # handle charts
    chart_builders = {
        "latency": build_latency_chart,
        "delay": build_delay_chart,
        "volume": build_volume_chart,
        "hosts_dcount": build_hosts_dcount_chart,
        "data_sampling_anomaly": build_data_sampling_anomaly_chart,
        "flx_status": build_flx_status_chart,
        "ml_outliers": build_ml_outliers_chart,
        "flx_metric": build_flx_generic_metric_chart,
        "wlk_metric": build_wlk_generic_metric_chart,
    }

    # get the comma separated list of recipients
    recipients = helper.get_param("email_recipients")
    # turn into a list
    recipients = recipients.split(",")

    # consider orange as in alerting state (0 for false, 1 for true)
    consider_orange_as_alerting_state = int(
        helper.get_param("orange_as_alerting_state")
    )

    # Determine alerting states based on configuration
    alerting_states = ["red"]
    if consider_orange_as_alerting_state:
        alerting_states.append("orange")

    # Get environment_name
    environment_name = helper.get_param("environment_name")
    if not environment_name or environment_name == "" or environment_name == "null":
        environment_name = "Splunk"

    # Get drilldown_root_uri
    drilldown_root_uri = helper.get_param("drilldown_root_uri")

    # get email_account_settings
    email_account_settings = {}

    endpoint = (
        f"{server_uri}/services/trackme/v2/configuration/get_emails_delivery_account"
    )
    try:
        response = requests.post(
            endpoint,
            headers=header,
            json={"account": email_account},
            verify=False,
            timeout=300,
        )
        response.raise_for_status()
        email_account_settings = response.json()
    except Exception as e:
        helper.log_error(
            f"Error getting email account settings: exception={str(e)}, email_account={email_account}"
        )
        return 1

    # get allowed_email_domains value (optional, comma separated list of allowed email domains)
    allowed_email_domain = email_account_settings.get("allowed_email_domains", None)

    #
    # Main: process upstream events
    #

    # Loop through the events
    for event in helper.get_events():

        # get tenant_id
        tenant_id = event["tenant_id"]

        # get the trackme_summary_idx for the tenant
        tenant_idx = trackme_idx_for_tenant(session_key, server_uri, tenant_id)
        tenant_trackme_summary_idx = tenant_idx["trackme_summary_idx"]

        # get alias, object, object_category, object_id
        alias = event.get("alias", None)
        object = event["object"]
        object_category = event["object_category"]

        # get component from object_category (splk-<component>)
        component = object_category.split("-")[1]

        # connect to the main tenant KVstore collection
        collection_main_name = f"kv_trackme_{component}_tenant_{tenant_id}"
        collection_main = service.kvstore[collection_main_name]

        # connect to the stateful KVstore collection
        collection_stateful_alerting_name = (
            f"kv_trackme_stateful_alerting_tenant_{tenant_id}"
        )
        collection_stateful_alerting = service.kvstore[
            collection_stateful_alerting_name
        ]

        # connect to the stateful KVstore collection for charts storing
        collection_stateful_alerting_charts_name = (
            f"kv_trackme_stateful_alerting_charts_tenant_{tenant_id}"
        )
        collection_stateful_alerting_charts = service.kvstore[
            collection_stateful_alerting_charts_name
        ]

        # get the drilldown_link
        drilldown_link = generate_drilldown_link(
            parsed_url, drilldown_root_uri, tenant_id, object_category, object
        )

        # get object_id, if not part of the upstream event, get it from the main KVstore
        object_id = event.get("keyid", None)
        if not object_id:
            object_id = get_keyid_from_main_kvstore(helper, collection_main, object)
        if not object_id:
            helper.log_error(
                f"tenant_id={tenant_id}, object={object}, object_category={object_category}, object_id=None, No object_id found for object"
            )
            continue

        # get the object_state from the main KVstore collection
        object_state, anomaly_reason, status_message = (
            get_object_state_from_main_kvstore(helper, collection_main, object_id)
        )
        helper.log_info(
            f"Processing event: tenant_id={tenant_id}, object={object}, object_category={object_category}, object_id={object_id}, object_state={object_state}"
        )

        # get the value of anomaly_reason, and ensure it is a list
        anomaly_reason = event.get("anomaly_reason", None)
        if anomaly_reason:
            if not isinstance(anomaly_reason, list):
                anomaly_reason = [anomaly_reason]

        # Determine which chart types to include based on anomaly_reason

        chart_types = []
        model_id = None  # for outliers

        # for dsm/dhm, add latency, delay and volume as the basis
        if component in ("dsm", "dhm"):
            chart_types = [("latency", None), ("delay", None), ("volume", None)]

        elif component == "flx":
            chart_types = [("flx_status", None)]

            # dynamic metrics
            flx_metrics = flx_get_metrics_catalog_for_object_id(
                helper, service, tenant_id, object_id
            )
            helper.log_debug(f"flx_metrics: {flx_metrics}")
            if flx_metrics:
                chart_types.append(("flx_metric_group", flx_metrics))  # Pass whole list

        elif component == "wlk":

            # dynamic metrics
            wlk_metrics = wlk_get_metrics_catalog_for_object_id(
                helper, service, tenant_id, object_id
            )
            helper.log_debug(f"wlk_metrics: {wlk_metrics}")
            if wlk_metrics:
                chart_types.append(("wlk_metric_group", wlk_metrics))  # Pass whole list

        # relavant for dsm only:
        if component == "dsm":
            if "min_hosts_dcount" in anomaly_reason:
                chart_types.append(("hosts_dcount", None))

            if "data_sampling_anomaly" in anomaly_reason:
                chart_types.append(("data_sampling_anomaly", None))

        # ML Outliers:
        # - try to retrieve the mlmodels_in_anomaly from the KVstore
        mlmodels_in_anomaly = get_mlmodels_in_anomaly_from_kvstore(
            helper, service, tenant_id, component, object, object_id
        )
        if mlmodels_in_anomaly:
            for model_id in mlmodels_in_anomaly:
                chart_types.append(("ml_outliers", model_id))

        #
        # stateful record lifecycle:
        # - an alert is closed when the object_state is set to "green"
        # - an alert is opened when the object_state is set to "red"
        # - if the alert is closed and the object_state is not green, this is a new alert and therefore a new message_id
        #

        # alert_new_thread boolean
        alert_new_thread = False

        # Generate a new email-specific message ID
        current_email_id = generate_message_id()

        # get the stateful record, if any
        stateful_record = get_stateful_records_for_object_id(
            helper, collection_stateful_alerting, object_id
        )

        # if the stateful record is not found, we need to create a new thread
        if not stateful_record:
            alert_new_thread = True
            incident_id = (
                generate_message_id()
            )  # Generate a stable ID for the lifecycle
        else:
            incident_id = stateful_record.get("incident_id")  # Reuse it if updating

        # check if the event should be skipped
        if stateful_record:

            # get the mtime from stateful_record, of it exists
            mtime = float(stateful_record.get("mtime"))

            # get the event time
            event_time = float(event.get("_time", time.time()))

            # if we have a stateful record, and the event time is not newer than the mtime, we can skip the processing
            if event_time <= mtime:
                helper.log_info(
                    f"Skipping event: tenant_id={tenant_id}, object={object}, object_id={object_id}, event_time={time.strftime('%c', time.localtime(event_time))}, mtime={time.strftime('%c', time.localtime(mtime))}, event is not newer than stateful record last update"
                )
                continue

        # log
        helper.log_info(
            f"Processing alert: tenant_id={tenant_id}, object={object}, object_id={object_id}, object_state={object_state}, alert_new_thread={alert_new_thread}"
        )

        # message_id, if this is a new thread, we need to generate a new message_id
        message_id = None

        if alert_new_thread:
            message_id = current_email_id
            in_reply_to = None
            reference_chain = []
        else:
            message_id = stateful_record.get("message_id")
            in_reply_to = message_id
            reference_chain = stateful_record.get("reference_chain", [message_id])
            if not isinstance(reference_chain, list):
                reference_chain = [message_id]
            reference_chain.append(current_email_id)

        # message
        message = None

        # message_source and id
        message_source = event.get("sourcetype")
        message_source_id = event.get("event_id")

        # trackme:flip provides the message as result
        if message_source == "trackme:flip":
            message = event["result"]

        elif message_source == "trackme:notable":
            message = event["status_message"]

        elif message_source == "trackme:sla_breaches":
            message = event["sla_message"]

        # do no process if no stateful_record and object_state is green/blue
        if not stateful_record and object_state not in alerting_states:
            helper.log_info(
                f"Skipping processing: tenant_id={tenant_id}, object={object}, object_id={object_id}, object_state={object_state}, event_id={event.get('event_id')}, no stateful record and non-alerting state"
            )
            continue

        # set delivery_type as a list depending on the delivery_target
        delivery_type = []
        if delivery_target in ("emails_and_ingest", "emails_only"):
            delivery_type.append("email")
        if delivery_target in ("ingest_only", "emails_and_ingest"):
            delivery_type.append("ingest")

        # init detection_time (human readable format %c %Z)
        detection_time = time.strftime(
            "%c %Z", time.localtime(float(event.get("_time", time.time())))
        )

        # init new_record
        new_record = {}

        # if necessary, insert the record
        if not stateful_record:

            # init messages
            messages = [message]

            record_key = hashlib.sha256(
                f"{object_id}{time.time()}".encode()
            ).hexdigest()
            new_record = {
                "_key": record_key,
                "tenant_id": tenant_id,
                "object": object,
                "object_category": object_category,
                "object_id": object_id,
                "object_state": object_state,
                "incident_id": incident_id,
                "message_id": current_email_id,
                "messages": messages,
                "reference_chain": reference_chain,
                "delivery_type": delivery_type,
                "alert_status": "opened",
                "ctime": float(event.get("_time", time.time())),
                "mtime": float(event.get("_time", time.time())),
            }
            insert_stateful_record(
                collection_stateful_alerting,
                new_record,
            )

        else:

            # handle message
            messages = stateful_record.get("messages", [])
            # if messages is not a list, convert it to a list and add the message
            if not isinstance(messages, list):
                messages = [messages]
            messages.append(message)

            # if object_state is not green, this is an update
            if object_state in alerting_states:
                new_record = {
                    "_key": stateful_record.get("_key"),
                    "tenant_id": tenant_id,
                    "object": object,
                    "object_category": object_category,
                    "object_id": object_id,
                    "object_state": object_state,
                    "incident_id": incident_id,
                    "message_id": current_email_id,
                    "messages": messages,
                    "reference_chain": reference_chain,
                    "delivery_type": delivery_type,
                    "alert_status": "updated",
                    "ctime": float(stateful_record.get("ctime")),
                    "mtime": float(event.get("_time", time.time())),
                }
                update_stateful_record(
                    collection_stateful_alerting,
                    new_record,
                )

            else:
                new_record = {
                    "_key": stateful_record.get("_key"),
                    "tenant_id": tenant_id,
                    "object": object,
                    "object_category": object_category,
                    "object_id": object_id,
                    "object_state": object_state,
                    "incident_id": incident_id,
                    "message_id": current_email_id,
                    "messages": messages,
                    "reference_chain": reference_chain,
                    "delivery_type": delivery_type,
                    "alert_status": "closed",
                    "ctime": float(stateful_record.get("ctime")),
                    "mtime": float(event.get("_time", time.time())),
                }
                update_stateful_record(
                    collection_stateful_alerting,
                    new_record,
                )

        #
        # Charts generation
        #

        chart_ids = []
        chart_descriptions = []

        if generate_charts:
            chart_results_map = {}
            search_kwargs = {
                "output_mode": "json",
                "count": 0,
                "earliest_time": "-24h@h",
                "latest_time": "now",
            }

            for chart_type, model_or_metrics in chart_types:
                metric_list = (
                    model_or_metrics if isinstance(model_or_metrics, list) else None
                )
                model_id = (
                    model_or_metrics if isinstance(model_or_metrics, str) else None
                )

                if chart_type in ("flx_metric_group", "wlk_metric_group"):
                    for metric_name in model_or_metrics:
                        if "flx" in chart_type:
                            chart_descriptions.append(f"FLX Metric: {metric_name}")
                        else:
                            chart_descriptions.append(f"WLK Metric: {metric_name}")
                elif chart_type == "ml_outliers":
                    chart_descriptions.append(f"ML Outlier: {model_or_metrics}")
                elif chart_type == "hosts_dcount":
                    chart_descriptions.append("Hosts DCount")
                elif chart_type == "latency":
                    chart_descriptions.append("Latency Metrics")
                elif chart_type == "delay":
                    chart_descriptions.append("Delay Metrics")
                elif chart_type == "volume":
                    chart_descriptions.append("Volume Metrics")
                elif chart_type == "data_sampling_anomaly":
                    chart_descriptions.append("Sampling Anomaly %")
                elif chart_type == "flx_status":
                    chart_descriptions.append("FLX Status")
                else:
                    chart_descriptions.append(chart_type.title())  # Fallback

                search_query = get_chart_search(
                    chart_type=chart_type,
                    tenant_id=tenant_id,
                    object_category=object_category,
                    object=object,
                    keyid=object_id,
                    model_id=model_id,
                    metric_list=metric_list,
                )

                if not search_query:
                    helper.log_info(
                        f"Skipping chart type: {chart_type}, no search query returned."
                    )
                    continue

                key = (
                    chart_type
                    if metric_list
                    else f"{chart_type}:{model_id}" if model_id else chart_type
                )

                helper.log_info(f"Running search: {search_query}")
                try:
                    reader = run_splunk_search(
                        service, search_query, search_kwargs, 24, 5
                    )
                    chart_results_map[key] = [
                        item for item in reader if isinstance(item, dict)
                    ]
                except Exception as e:
                    helper.log_error(f"Search failed for chart {key}, error: {e}")

            # Build charts from the gathered results
            chart_svg_base64_list = build_requested_charts(
                helper,
                chart_builders,
                chart_types,
                chart_results_map,
                theme_charts=theme_charts,
            )

            # Store charts to KVstore
            if chart_svg_base64_list:
                try:
                    for idx, svg_b64 in enumerate(chart_svg_base64_list):
                        chart_id = hashlib.sha256(
                            f"{incident_id}-{idx}-{time.time()}".encode()
                        ).hexdigest()
                        chart_record = {
                            "_key": chart_id,
                            "object_category": object_category,
                            "object": object,
                            "object_id": object_id,
                            "incident_id": incident_id,
                            "message_id": current_email_id,
                            "ctime": float(time.time()),
                            "chart_id": chart_id,
                            "chart_description": (
                                chart_descriptions[idx]
                                if idx < len(chart_descriptions)
                                else f"Chart {idx + 1}"
                            ),
                            "chart_svg_base64": svg_b64,
                        }
                        collection_stateful_alerting_charts.data.insert(
                            json.dumps(chart_record)
                        )
                        chart_ids.append(chart_id)
                    helper.log_info(
                        f"{len(chart_svg_base64_list)} charts stored in KVstore for incident_id={incident_id} with chart_ids={chart_ids}"
                    )
                except Exception as e:
                    helper.log_error(
                        f"Failed to store charts for incident_id={incident_id}: {e}"
                    )

        #
        # Ingest stateful alert event into the TrackMe summary index
        #

        if delivery_target in ("emails_and_ingest", "emails_only"):

            # calculate the event_id as the sha-256 sum of the record
            event_id = hashlib.sha256(json.dumps(new_record).encode()).hexdigest()

            # add the event_id to the record
            new_record["event_id"] = event_id

            # add detection_time to the record
            new_record["detection_time"] = detection_time

            # remove the _key from the record
            new_record.pop("_key", None)

            # include chart_ids in the record
            new_record["chart_ids"] = chart_ids

            # include drilldown_link in the record
            new_record["drilldown_link"] = drilldown_link

            # include priority in the record
            new_record["priority"] = event.get("priority", None)

            # include message_source and message_source_id in the record
            new_record["message_source"] = message_source
            new_record["message_source_id"] = message_source_id

            try:
                ingest_stateful_alert_event(
                    helper,
                    session_key,
                    server_uri,
                    tenant_id,
                    tenant_trackme_summary_idx,
                    new_record,
                )

            except Exception as e:
                helper.log_error(
                    f"Error ingesting stateful alert event: tenant_id={tenant_id}, object={object}, object_id={object_id}, error={e}"
                )

        #
        # Emails delivery
        #

        if delivery_target in ("emails_and_ingest", "emails_only"):

            # Loop through all provided recipients
            for recipient in recipients:
                recipient = recipient.strip()
                if not recipient:
                    continue

                # Build the subject prefix
                subject_prefix = f"[TrackMe Incident-ID: {incident_id}]"

                # Compose subject based on state
                if alert_new_thread:
                    subject_suffix = f" Entity {object} is now in alert"
                elif object_state not in alerting_states:
                    subject_suffix = (
                        f" Entity {object} has received an incident closure update"
                    )
                else:
                    subject_suffix = f" Entity {object} has received an update"

                # Final subject
                subject = f"{subject_prefix}{subject_suffix}"

                # check if the recipient is allowed
                if allowed_email_domain:
                    if not any(
                        domain in recipient
                        for domain in allowed_email_domain.split(",")
                    ):
                        helper.log_warn(
                            f"Skipping recipient: tenant_id={tenant_id}, object={object}, object_id={object_id}, recipient={recipient}, not in allowed email domains"
                        )
                        continue

                # get email format
                email_format = email_account_settings.get("email_format", "html")

                # transition message
                if object_state in alerting_states:
                    if alert_new_thread:
                        transition_message = (
                            "🚨 The entity has transitioned to an alerting state:"
                            if email_format == "html"
                            else "The entity has transitioned to an alerting state:"
                        )
                    else:
                        transition_message = (
                            "ℹ️ The entity has received an update:"
                            if email_format == "html"
                            else "The entity has received an update:"
                        )
                else:
                    transition_message = (
                        "✅ The entity has received an incident closure update and is now in a non-alerting state:"
                        if email_format == "html"
                        else "The entity has received an incident closure update and is now in a non-alerting state:"
                    )

                #
                # Generate and send email
                #

                try:
                    msg_id = send_trackme_email(
                        helper=helper,
                        config=email_account_settings,
                        message_id=current_email_id,
                        to_address=recipient,
                        subject=subject,
                        core_message=message,
                        in_reply_to=in_reply_to,
                        reference_chain=reference_chain,
                        format=email_format,
                        tenant_id=tenant_id,
                        alias=alias,
                        object=object,
                        object_state=object_state,
                        object_category=object_category,
                        object_id=object_id,
                        anomaly_reason=anomaly_reason,
                        status_message=status_message,
                        message_source=message_source,
                        message_source_id=message_source_id,
                        transition_message=transition_message,
                        environment_name=environment_name,
                        drilldown_link=drilldown_link,
                        incident_id=incident_id,
                        detection_time=detection_time,
                        chart_svg_base64_list=chart_svg_base64_list,
                    )
                    helper.log_info(
                        f"Email sent: tenant_id={tenant_id}, object={object}, object_id={object_id}, recipient={recipient}, message_id={msg_id}"
                    )

                except Exception as e:
                    helper.log_error(
                        f"Error sending email: tenant_id={tenant_id}, object={object}, object_id={object_id}, recipient={recipient}, error={e}"
                    )

    #
    # End of processing
    #

    return 0
